/** Copyright (C) 2008 Tom Fogal
 * simplified BSD licensed I guess.
 * I'm assuming that's compatible with the existing IceT license .. */
#include <assert.h>
#include <inttypes.h>
#include <string.h>

#include <GL/ice-t.h>
#include <image.h>
#include <state.h>
#include <context.h>
#include <diagnostics.h>
#include "common.h"

#include <glib.h>
#include <mpi.h>
#if 0
#   include <mpe.h>
#endif
#include "coord-paic.h"
#include "paic-collab.h"
#include "common-paic.h"

static MPI_Request *r_ready; /* requests to wait for ready. */
static MPI_Request mr_buddy; /* request for a compositing buddy */
static size_t cur_buddy;     /* current choice for a buddy to composite w/ */

enum PAIC_Tags {
    TAG_RENDERED,         /* local render is complete msg */
    TAG_COMPOSITE_REQUEST /* will you composite with me? msg */
};

static IceTImage paic(void);
IceTStrategy ICET_STRATEGY_PAIC = { "PAIC", ICET_FALSE, paic };

/** Have we initialized our data structures yet? */
static gboolean initialized = 0;

static void initialize();

#if 0
#   define mpe(block) do { block } while(0)
#else
#   define mpe(block) /* no-op */
#endif

#define TIME_BLOCK_STR(str_pre, str_post, block)         \
    do {                                                 \
        int _head, _tail;                                \
        mpe(MPE_Log_get_state_eventIDs(&_head, &_tail)); \
        mpe(MPE_Log_event(_head, 1, str_pre));           \
        block                                            \
        mpe(MPE_Log_event(_tail, 2, str_post));          \
    } while(0)

gpointer f_render_tile(gpointer data);

/** Entry point for the PAIC. */
static IceTImage
paic(void)
{
    IceTImage image;
    IceTImage img_buffer;
    IceTSparseImage img_in, img_out;
    GLint max_pixels;
    GLint *compose_group;
    GLint *display_nodes;
    GThread *asynch_render;

    icetRaiseDebug("In PAIC!");
    if(initialized == FALSE) {
        initialize();
    }
    assert(initialized == TRUE);

    size_t tile = 0;
    assert(pk_tiles() == 1); /* current HACK -- assume one tile */
    /* create a thread to render in the bg */
    asynch_render = g_thread_create(f_render_tile, &tile, TRUE, NULL);

    icetGetIntegerv(ICET_TILE_MAX_PIXELS, &max_pixels);
    display_nodes = icetUnsafeStateGet(ICET_DISPLAY_NODES);

    icetResizeBuffer(icetFullImageSize(max_pixels)*2   +
                     icetSparseImageSize(max_pixels)*2 +
                     sizeof(int)*pk_procs());

    image = NULL;
    img_buffer = icetReserveBufferMem(icetFullImageSize(max_pixels));
    img_in  = icetReserveBufferMem(icetSparseImageSize(max_pixels));
    img_out = icetReserveBufferMem(icetSparseImageSize(max_pixels));
    compose_group = icetReserveBufferMem(sizeof(GLint)*pk_procs());

    for(size_t i=0; i < pk_procs(); ++i) {
        compose_group[i] = i;
    }

    mpe(MPE_Init_log());
    mpe(MPE_Start_log());

    /* Where should the image go?  Depends if I'm a display node .. */
    IceTImage ibuf = img_buffer;
    if(display_nodes[0] == pk_rank()) {
        image = icetReserveBufferMem(icetFullImageSize(max_pixels));
        ibuf = image;
    }

    g_thread_join(asynch_render);
    /* coordinate the composition process. */
    coord_render_complete(); /* notify all nodes we're done. */
    int buddy = -1;
    do {
        TIME_BLOCK_STR("pre-coord", "post-coord",
            buddy = coord_master();
        );
    } while(buddy != COORD_COMPLETE);

    hcRead(tile, ibuf);
#if 0
        icetBswapCompose(compose_group, pk_procs(), image_dest,
                         ibuf, img_in, img_out);
#endif
    MPI_Barrier(MPI_COMM_WORLD);
    mpe(MPE_Stop_log());
    MPI_Barrier(MPI_COMM_WORLD);
    mpe(MPE_Finish_log("rendertime"));

    return image;
}

static void
initialize()
{
    g_thread_init(NULL);
    r_ready = malloc(sizeof(MPI_Request)*pk_procs());
    initialized = TRUE;
}

/** renders the given index; asynch entry point. */
gpointer
f_render_tile(gpointer data)
{
    size_t *idx = (size_t *)data;
    TIME_BLOCK_STR("pre-hcRender", "post-hcRender",
        hcRender(*idx);
    );
    free(idx);
    return NULL;
}

#if 0
gpointer
thr_status_updates(gpointer data)
{
    MPI_Request *fromwho;
    fromwho = g_malloc0(sizeof(int) * pk_procs());
    for(size_t i = 0; i < pk_procs(); ++i) {
        fromwho[i] = i;
    }
    do {
        int recvbuf[2];
        int idx;
        MPI_Status status;
        MPI_Waitany(2, recvbuf, &idx, &status);
    } while(1);
}
#endif
